#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2013 Netflix, Inc.


"""Command line tools entry point."""

from cligraphy.core import capture, decorators, read_configuration, ctx
from cligraphy.core.log import setup_logging
from cligraphy.core.parsers import AutoDiscoveryCommandMap, SmartCommandMapParser, ParserError, CustomDescriptionFormatter
from cligraphy.core.util import undecorate_func, pdb_wrapper, profiling_wrapper, call_chain

from remember.memoize import memoize

import faulthandler
from setproctitle import setproctitle  # pylint:disable=no-name-in-module

import argcomplete
import argparse
import logging
import logging.config
import os
import signal
import subprocess
import sys


def _warn_about_bad_non_ascii_chars(args):
    """Detect non-ascii variants of some interesting characters, such as — instead of -"""
    bad_chars = (
        u'—',
        u'…',
        u'“',
        u'”',
        u'\u200b',  # zero-width space
    )
    try:
        line = ' '.join(arg.decode(sys.stdout.encoding or 'UTF-8') for arg in args)
        bad = [ char in bad_chars for char in line ]
        if any(bad):
            logging.warning('Your command line contains %d bad unicode character(s), did you copy/paste from a tool that garbles text?', len(bad))
            logging.warning('> ' + line)
            logging.warning('> ' + ''.join('^' if isbad else ' ' for isbad in bad))
            return
    except Exception:  # pylint:disable=broad-except
        logging.warning('Exception while trying to detect bad unicode characters in command line args; continuing...', exc_info=True)


def _warn_about_bad_path(env_root, path):
    """Warn about mistakes in $PATH"""
    if not env_root:
        logging.warning('Running oc outside of its virtualenv is supported, but untested. Please report any bugs!')
        return
    env_bin = os.path.join(env_root, 'bin')
    elements = path.split(':')
    try:
        position = elements.index(env_bin)
        if position != 0:
            logging.warning('Your oc virtualenv, %s, is not the first element of your $PATH. Things might be broken.', env_root)
    except ValueError:
        logging.warning('Your oc virtualenv, %s, is not listed in your $PATH. Things are likely broken.', env_root)
        return



class _VersionAction(argparse.Action):
    """Shows last commit information"""
    def __call__(self, *args, **kwargs):
        try:
            os.chdir(ctx.conf.tool.repo_path)
            last_commit = subprocess.check_output(['git', 'log', '-1'], stderr=subprocess.PIPE)
        except (subprocess.CalledProcessError, OSError):
            last_commit = '(could not get more recent commit information)'
        print '%s v%s\n%s' % (ctx.conf.tool.name, ctx.conf.tool.version, last_commit)
        sys.exit(1)


class Cligraph(object):

    def __init__(self, name, shortname, path):
        assert name
        assert shortname
        assert path
        self.tool_name = name
        self.tool_shortname = shortname
        self.tool_path = path
        self.conf, self.conf_layers = read_configuration(self)
        ctx.cligraph = self


    # FIXME(stf/oss): header names for octools
    def _setup_requests_audit_headers(self, command):
        """Setup requests user-agent and x-user/x-app headers.
        Just best effort - we don't care that much if this fails"""
        try:
            import requests
            def _default_user_agent(*args):
                return 'requests (cligraphy/%s)' % command
            requests.utils.default_user_agent = _default_user_agent

            base_default_headers = requests.utils.default_headers
            def _default_headers(*args):
                headers = base_default_headers()
                headers['X-User'] = os.getenv('USER')
                headers['X-App'] = 'cligraphy/%s' % command
                return headers
            requests.utils.default_headers = _default_headers
            requests.sessions.default_headers = _default_headers
        except:
            logging.warn("Could set up requests audit headers, continuing anyway")
            pass

    # pylint:disable=protected-access
    def _run(self, args):
        """Run command by calling the main() function correctly (how we pass args depends on its actual signature)."""
        # if the main method has been decorated, we need to look at the original function's argspec (but call the wrapper)
        import inspect
        orig_func, _ = undecorate_func(args._func)
        argspec = inspect.getargspec(orig_func)
        if len(argspec.args) == 0:
            return args._func()
        elif argspec.varargs or argspec.keywords or len(argspec.args) > 1:
            kwargs = dict(**vars(args))
            for kw in kwargs.keys():
                if kw.startswith('_'):
                    logging.debug('removing internal arg %s', kw)
                    del kwargs[kw]
            return args._func(**kwargs)
        else:
            if argspec.args[0] != 'args':
                raise Exception('Programming error in command: if main() only has one argument it must be called "args"')
            func = args._func
            for kw in vars(args).keys():
                if kw.startswith('_') and kw not in ('_parser', '_cligraph'):  #FIXME(stf/oss) maybe just expose cligraph
                    logging.debug('removing internal arg %s', kw)
                    delattr(args, kw)
            return func(args)


    def _run_command_process(self, args):
        """Command (child) process entry point. args contains the function to execute and all arguments."""

        setup_logging(args._level)

        command = ' '.join(sys.argv[1:])
        setproctitle('oc/command/%s' % command)
        faulthandler.register(signal.SIGUSR2, all_threads=True, chain=False)  # pylint:disable=no-member
        self._setup_requests_audit_headers(command)

        ret = 1
        try:
            chain = [self._run]
            if args._profile:
                chain.append(profiling_wrapper)
            if args._pdb:
                chain.append(pdb_wrapper)
            ret = call_chain(chain, args)
        except SystemExit as exc:
            ret = exc.code
        except ParserError as pe:
            pe.report()
        except Exception:  # pylint:disable=broad-except
            logging.exception('Top level exception in command process')
        finally:
            sys.exit(ret)


    def _parse_args(self):
        """Setup parser and parse cli arguments.
        NB! counter-intuitively, this function also messes around with logging levels.
        """

        # We want some of our options to take effect as early as possible, as they affect command line parsing.
        # For these options we resort to some ugly, basic argv spotting

        if '--debug' in sys.argv:
            logging.getLogger().setLevel(logging.DEBUG)
            logging.debug('Early debug enabled')

        if '--verbose' in sys.argv or '-v' in sys.argv:
            logging.getLogger().setLevel(logging.INFO)

        autodiscover = False
        if '--autodiscover' in sys.argv:
            logging.debug('Autodiscover enabled')
            autodiscover = True

        parser = SmartCommandMapParser(prog="oc",
                                       description="Cligraphy command line tools",
                                       formatter_class=CustomDescriptionFormatter)

        self.parser = parser  # expose to eg. ctx

        parser.add_argument('--version', action=_VersionAction, nargs=0, dest="_version")
        parser.add_argument("--debug", help="enable debuging output", dest="_level", action="store_const", const=logging.DEBUG)
        parser.add_argument("--pdb", help="run pdb on exceptions", dest="_pdb", action="store_true")
        parser.add_argument("--no-capture", help="disable input/output capture", dest="_capture", action="store_false", default=True)
        parser.add_argument("--no-reporting", help="disable reporting", dest="_reporting", action="store_false", default=True)
        parser.add_argument("--profile", help="enable profiling", dest="_profile", action="store_true", default=False)
        parser.add_argument("--autodiscover", help="re-discover commands and refresh cache (default: read cached commands list)", dest="_autodiscover", action="store_true")
        parser.add_argument("-v", "--verbose", help="enable informational output", dest="_level", action="store_const", const=logging.INFO)

        for namespace, command_map in self.get_command_maps(autodiscover):
            parser.add_command_map(namespace, command_map)

        argcomplete.autocomplete(parser)

        _warn_about_bad_non_ascii_chars(sys.argv)
        _warn_about_bad_path(os.getenv('VIRTUAL_ENV'), os.getenv('PATH'))

        args = parser.parse_args()
        args._parser = parser  # deprecated

        # pylint:disable=protected-access
        if args._level is not None:
            logging.getLogger().setLevel(args._level)

        return args

    @memoize()
    def get_command_maps(self, autodiscover=False):
        """Get all the command maps defined in our configuration.

        If autodiscover is True (defaults to False), python commands will be autodiscovered (instead of simply being obtained
        from a cached command map)."""

        result = []
        for module, options in self.conf.commands.items():
            try:
                if options is None:
                    options = {}
                logging.debug('Configuring commands module %s with options %s', module, options)

                opt_type = options.get('type', 'python')
                opt_namespace = options.get('namespace', '')

                if opt_type == 'python':
                    result.append((opt_namespace, AutoDiscoveryCommandMap(self, module).build(force_autodiscover=autodiscover)))
                else:
                    raise Exception('Dont know how to handle commands module with type [%s]', opt_type)
            except Exception as exc:  # pylint:disable=broad-except
                logging.warning('Could not configure commands module [%s] defined in configuration: %s. Skipping it.', module, exc,
                                exc_info=True)

        return result

    def main(self):
        """Main oc wrapper entry point."""

        setup_logging()

        try:
            args = self._parse_args()
            logging.debug("Parsed args: %r", vars(args))
        except ParserError as pe:
            pe.report()

        if not os.isatty(sys.stdin.fileno()) or not os.isatty(sys.stdout.fileno()):
            logging.info('stdin or stdout is not a tty, disabling capture')
            args._capture = False

        # reporting: we send command executions report to a web service; unless the report.enabled conf key is false
        from cligraphy.core.reporting import NoopReporter, ToolsPadReporter
        reporter = ToolsPadReporter(self) if (self.conf.report.enabled and args._reporting) else NoopReporter()
        reporter.report_command_start(sys.argv)

        decs = decorators.get_tags(args._func)
        logging.debug("Decorator tags: %s", decs)

        if decorators.Tag.interactive in decs or not args._capture:
            recorder = capture.NoopOutputRecorder()
        else:
            recorder = capture.BufferingOutputRecorder(max_output_size=self.conf.report.max_output_size)

        setproctitle('oc/parent/%s' % ' '.join(sys.argv[1:]))
        faulthandler.register(signal.SIGUSR2, all_threads=True, chain=False)  # pylint:disable=no-member

        if args._capture:
            # go ahead and run out command in a child process, recording all I/O
            logging.debug('Parent process %d ready to execute command process', os.getpid())
            status = capture.spawn_and_record(recorder, self._run_command_process, reporter.start, args)
            logging.debug('Command process exited with status %r', status)
        else:
            reporter.start()
            try:
                self._run_command_process(args)
            except SystemExit as exc:
                status = exc.code

        # report execution details
        reporter.report_command_exit(status)
        reporter.report_command_output(recorder.output_as_string())  # TODO(stefan): also report total output size
        reporter.stop()
        return status



#!/usr/bin/env python
# Copyright 2013, 2014 Netflix, Inc.

"""Command line tools entry point"""

from cligraphy.core.util import try_import

from remember import memoize

import json
import logging
import logging.config
import os
import sys


class JsonHandler(logging.Handler):

    def __init__(self, destination, *args, **kwargs):
        self.destination = destination
        self.recurse = False
        super(JsonHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        if self.recurse:
            return
        try:
            self.recurse = True
            self._emit(record)
        finally:
            self.recurse = False

    def _emit(self, record, first_attempt=True):
        try:
            msg = record.msg % record.args
        except TypeError:
            msg = str(record.msg) + ' (formatting error)'

        output = {}
        try:
            output = {
                'ts': int(record.created * 1000),
                'level': record.levelno,
                'msg': msg,

                # module, function, ...
                'module': record.module,
                'file': record.filename,
                'func': record.funcName,
                'line': record.lineno,

                # process and thread information
                'tid': record.thread,
                'tname': record.threadName,
                'pid': record.process,
                'pname': record.processName,
            }

            if hasattr(record, 'extra'):
                if first_attempt:
                    output['extra'] = record.extra
                else:
                    output['extra'] = {'_omitted': True}  # We tried to serialize our record once and got an error, so omit ctx now

            if record.exc_info and record.exc_info[0] and record.exc_info[1]:
                output['exc_type'] = record.exc_info[0].__name__
                output['exc_msg'] = repr(record.exc_info[1].message)
                tb = []
                cur = record.exc_info[2]
                while cur:
                    frame = cur.tb_frame
                    tb.append((frame.f_code.co_filename, frame.f_code.co_name, frame.f_lineno))
                    cur = cur.tb_next
                output['exc_tb'] = tb

            try:
                msg = json.dumps(output)
                self.destination.log_event('log', record.created, msg)
            except (TypeError, OverflowError):
                if first_attempt:
                    return self.emit(record, False)  # Try again, without ctx this time
                else:
                    raise

        except (KeyboardInterrupt, SystemExit):
            raise
        except:  # pylint:disable=bare-except
            print 'log.structured: serialization error - details follow'
            print(output)
            self.handleError(record)


def silence_verbose_loggers():
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARN)


@memoize.memoize()
def _get_dict_config():
    dict_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'datefmt': '%Y-%m-%d %H:%M:%S',
                'format': "%(asctime)s %(levelname)-8s %(module)s %(funcName)s %(message)s"
            },
        },
        'handlers': {
            'defaultHandler': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'default'
            },
        },
        'loggers': {
            '': {
                'handlers': (['defaultHandler']),
                'level': 'WARN',
            },
        }
    }

    if os.isatty(sys.stderr.fileno()) and try_import('colorlog')[0]:
        dict_config['formatters']['colors'] = {
            '()': 'colorlog.ColoredFormatter',
            'datefmt': '%Y-%m-%d %H:%M:%S',
            'format': "%(log_color)s%(asctime)s %(levelname)-8s%(reset)s %(purple)s%(module)s/%(funcName)s%(reset)s %(message)s"
        }
        dict_config['handlers']['defaultHandler']['formatter'] = 'colors'

    return dict_config


def setup_logging(level=None):
    """Configure logging"""
    try:
        logging.config.dictConfig(_get_dict_config())
        logging.captureWarnings(True)
        silence_verbose_loggers()
        if level is not None:
            logging.getLogger().setLevel(level)
    except Exception:  # pylint:disable=broad-except
        logging.basicConfig(level=logging.WARN)
        logging.warn('Could not configure logging, using basicConfig', exc_info=True)



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Session and execution tracking"""


from collections import namedtuple
import uuid
import os

TrackingInformation = namedtuple('TrackingInformation', 'session_uuid,execution_uuid')


def get_tracking():
    """Initializes our tracking information"""
    session_uuid = os.getenv('CLIGRAPHY_SESSION_UUID', None)
    if session_uuid is None:
        # Create new session
        session_uuid = str(uuid.uuid4())
        os.environ['CLIGRAPHY_SESSION_UUID'] = session_uuid

    return TrackingInformation(session_uuid=session_uuid, execution_uuid=str(uuid.uuid4()))

TRACKING = get_tracking()



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""
Cligraphy tools
"""

from attrdict import AttrDict

import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import collections
import os
import os.path
import re
import logging


class Context(object):

    def __init__(self):
        self._cligraph = None

    @property
    def cligraph(self):
        return self._cligraph

    @cligraph.setter
    def cligraph(self, value):
        self._cligraph = value

    @property
    def conf(self):
        return self._cligraph.conf

    @property
    def parser(self):
        return self._cligraph.parser



ctx = Context()


__CONF_SUBST_RE = re.compile(r'(%cfg.([a-zA-Z0-9.-_]+)%)')


def dictify_recursive(obj):
    """Transforms an attrdict tree into regular python dicts"""
    for key, val in obj.iteritems():
        if isinstance(val, AttrDict):
            obj[key] = dictify_recursive(val)
    return dict(**obj)


def update_recursive(base, overlay):
    """Resursively update base with values from overlay
    """
    for key, val in overlay.iteritems():
        if isinstance(val, collections.Mapping):
            base[key] = update_recursive(base.get(key, {}), val)
        else:
            base[key] = overlay[key]
    return base


def find_node(node, path, add=False):
    """Find a node in a dict tree. Path is the broken down path (array). Add missing nodes if add=True."""
    for part in path:
        if not node or not part in node:
            if add:
                node[part] = {}
            else:
                return None
        node = node[part]
    return node


def get(root, confkey):
    """Config key getter"""
    return find_node(root, confkey.split('.'))


def _resolve_config(root, subst_re, node=None):
    """recursive helper for resolve_config"""
    if node is None:
        node = root
    remaining = []
    for key, val in node.iteritems():
        if isinstance(val, collections.Mapping):
            remaining.extend(_resolve_config(root, subst_re, node=val))
        elif isinstance(val, basestring):
            for match, confkey in subst_re.findall(val):
                confval = get(root, confkey)
                if confval is not None and (not isinstance(confval, basestring) or subst_re.search(confval) is None):
                    node[key] = val = val.replace(match, str(confval))
                else:
                    remaining.append('%s: %s' % (key, val))
    return remaining


def resolve_config(cfg):
    """performs variable substitution in a configuration tree"""
    remaining = []
    for _ in range(16):  # 16 maximum substitution passes
        remaining_prev = remaining
        remaining = _resolve_config(cfg, __CONF_SUBST_RE)
        if not remaining or len(remaining) == len(remaining_prev):
            break
    if remaining:
        raise Exception('Incorrect configuration file: could not resolve some configuration variables:  %s' % remaining)


def automatic_configuration(cligraph, layer_name):
    auto_data = {
        'tool': {
            'name': cligraph.tool_name,
            'shortname': cligraph.tool_shortname,
            'version': '1.0',
            'repo_path': cligraph.tool_path,
        },
        'repos': {
            'root': os.path.abspath(os.path.join(cligraph.tool_path, '..')),
        },
    }
    username = os.getenv('USER')
    if username:
        auto_data['user'] = {
            'name': username,
            'email': '%s@domain.net' % (username)  # OPEN SOURCE TODO
        }
    else:
        auto_data['user'] = {
            'name': 'unknown',
            'email': 'octools-unknown-user@domain.net'  # OPEN SOURCE TODO
        }
    auto_data['user']['dotdir'] = os.path.abspath(os.path.expanduser('~/.' + cligraph.tool_shortname))
    return auto_data


def read_configuration(cligraph, custom_suffix=''):
    """Read configuration dict for the given tool
    """

    cfg = {}
    layers = collections.OrderedDict()
    layers['auto'] = [automatic_configuration, None]
    layers['shared'] = [os.path.join(cligraph.tool_path, 'conf/%s.yaml' % cligraph.tool_shortname), None]
    layers['custom'] = [os.path.join(os.path.abspath(os.path.expanduser('~/.' + cligraph.tool_shortname)), '%s.yaml%s' % (cligraph.tool_shortname, custom_suffix)), None]

    for layer_name, layer_data in layers.items():
        if callable(layer_data[0]):
            layer = layer_data[0](cligraph, layer_name)
        else:
            if not os.path.exists(layer_data[0]):
                continue
            with open(layer_data[0], 'r') as filep:
                layer = yaml.load(filep, Loader=Loader)
        layers[layer_name][1] = layer
        if layer:
            update_recursive(cfg, layer)

    resolve_config(cfg)
    return AttrDict(**cfg), layers


def write_configuration_file(filename, conf):
    """Write a config dict to the specified file, in yaml format"""
    import shutil

    try:
        shutil.copy2(filename, '%s.back' % filename)
    except IOError:
        pass

    try:
        os.makedirs(os.path.dirname(filename))
    except OSError:
        pass

    with open(filename, 'w') as filep:
        yaml.dump(conf, filep, indent=4, default_flow_style=False, Dumper=Dumper)


def edit_configuration(tool_name, callback):
    """Wrapper for configuration editing tools
    """
    import shutil

    filename = os.path.join(USER_DOTDIR, '%s.yaml' % tool_name)
    edit_filename = '%s.edit' % filename

    if os.path.exists(edit_filename):
        logging.warn('Unfinished edit file %s exists, editing that one.', edit_filename)
    else:
        if os.path.exists(filename):
            shutil.copy2(filename, edit_filename)
        else:
            with open(edit_filename, 'w') as fp:
                fp.write('\n')

    callback(edit_filename)

    read_configuration(tool_name, custom_suffix='.edit')  # validate that configuration is still readable

    if os.path.exists(filename):
        shutil.copy2(filename, '%s.back' % filename)
    os.rename(edit_filename, filename)



#!/usr/bin/env python
# Copyright 2013 Netflix, Inc.

"""Decorators."""

import cligraphy.core.parsers
from cligraphy.core.util import undecorate_func

from functools import wraps
import logging

from enum import Enum


class Tag(Enum):
    """Enumeration of the decorator tags we support."""
    beta = 1
    disabled = 2
    strict = 3
    interactive = 4


class DisabledCommandException(Exception):
    """Raised when a @disabled command is invoked."""
    pass


def tag(tag_enum, reason='No reason given'):
    """Decorator that disables a command."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
        wrapper.tag = tag_enum
        wrapper.original_func = func
        return wrapper
    return actual_decorator


def disabled(reason='No reason given'):
    """Decorator that disables a command."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            raise DisabledCommandException('This command is disabled: %s' % reason)
        wrapper.tag = Tag.disabled
        wrapper.original_func = func
        return wrapper
    return actual_decorator


def beta(reason='No reason given'):
    """Decorator for unstable commands - user will be prompted to confirm before execution."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            message = 'WARNING - this command is marked as beta: %s - are you sure you want to continue (y/N)? ' % reason
            logging.warning(message)
            confirm = raw_input(message)
            if confirm and confirm.lower() in ('y', 'yes'):
                func(*args, **kwargs)
            else:
                print 'Cancelled'
        wrapper.tag = Tag.beta
        wrapper.original_func = func
        return wrapper
    return actual_decorator


def strict(reason='No reason given'):
    """Command decorator that disables fuzzy matching."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if cligraphy.core.parsers.FUZZY_PARSED:
                logging.error('This command must be called by its exact name (%s). Please run as %s',
                              reason, cligraphy.core.parsers.FUZZY_PARSED[0][-1])
                return
            else:
                func(*args, **kwargs)
        wrapper.tag = Tag.strict
        wrapper.original_func = func
        return wrapper
    return actual_decorator


def get_tags(decorated_func):
    """Return the set of decorator tags applied to decorated_func."""
    _, decorators = undecorate_func(decorated_func)
    return set([ x.tag for x in decorators ])



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Human-friendly time period parsing
"""

import time
import math
from datetime import datetime
from dateutil import parser

UNITS = {
    'month': 3600*24*31,  # FIXME 1 month is not really 31 days
    'week': 3600*24*7,
    'day': 3600*24,
    'hour': 3600,
    'minute': 60,
    'second': 1,
}

UNITS['mon'] = UNITS['month']
UNITS['w'] = UNITS['week']
UNITS['d'] = UNITS['day']
UNITS['h'] = UNITS['hour']
UNITS['min'] = UNITS['minute']
UNITS['m'] = UNITS['minute']
UNITS['sec'] = UNITS['second']
UNITS['s'] = UNITS['second']

def parse_time(human_time):
    """Parse a human-friendly time expression, such as 1d or now, and return a unix timestamp
    """
    human_time = human_time.lower()
    now = int(time.time())
    if human_time == 'now':
        return now
    try:
        unit = human_time[-1]
        value = int(human_time[:-1])
        return now - (value * UNITS[unit])
    except:
        raise Exception('Dont know what to do with time [%s]' % human_time)

def parse_datetime(human_time):
    """Parse a human-friendly time expression (such as 1d or now)
            or datetime formatted as string (such as 2016-01-20 20:00)
        and return a datetime object
    """
    try:
        return datetime.fromtimestamp(parse_time(human_time))
    except:
        try:
            dt = parser.parse(human_time)
            return dt
        except:
            raise Exception('Dont know what to do with time [%s]' % human_time)

def align_times(start, end, ratio=0.02):
    """Align timestamps on a reasonable boundary
    """
    assert start < end, 'Start time must be before end time'
    boundary = int(math.ceil((end - start) * ratio))
    return int(start / boundary) * boundary, int(end / (boundary + 1)) * (boundary + 1)



#!/usr/bin/env python
# Copyright 2015, 2016 Netflix, Inc.

import logging
import time
import os.path

from decorator import decorator
from pathlib import Path


def retry(retry_count, exceptions, log_message, retry_sleep=0, backoff=1, maxdelay=None):
    """ Decorator to implement retry logic
        :retry_count int: number of attempts before aborting on failure.
        :exceptions tuple: exception classes to trap and retry on.
        :retry_sleep float or int: sets the initial delay between attempts in seconds.
        :backoff  float or int: sets the factor by which the delay should lengthen after each failure.
                                backoff must be greater than 1, or else it isn't really a backoff.
        :maxdelay int: exponential backoffs can get pretty lengthy. This limits the maximum delay

    """
    @decorator
    def _retry(f, *args, **kwargs):
        _tries, _delay = retry_count, retry_sleep
        while _tries > 0:
            try:
                return f(*args, **kwargs)
            except (exceptions) as e:
                logging.info('Failed to %s. Attempt: %s/%s', log_message, retry_count + 1 - _tries, retry_count)
                _tries -= 1
                if _tries == 0:
                    logging.error('Failed to %s with %s attempts', log_message, retry_count)
                    raise(e)
                time.sleep(_delay)
                _delay *= backoff
                if maxdelay and _delay > maxdelay:
                    _delay = maxdelay
    return _retry


def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i + n]


def is_file(parser, item):
    """
    Takes an argparse.ArgumentParser instance and value that you want
    validated.  If it is a file, returns a resolved pathlib.PosixPath object.
    If it isn't, prints the appropriate error and aborts execution via the
    ArguementParser.

    :param parser: ArgumentParser that you are trying to validate the item from
    :type parser: argparse.ArgumentParser
    :param item: path to file to test if is a file
    :type item: str
    :return: resolved Path object
    :rtype: pathlib.PosixPath
    """
    f = Path(os.path.expanduser(item))
    try:
        f = f.resolve()
    except IOError:
        parser.error('The file {file!r} does not exist'.format(file=item))
    if not f.is_file():
        parser.error('{item!r} is not a file'.format(item=item))
    try:
        with f.open('r') as fp:
            pass
    except IOError as e:
        parser.error('{item!r} cannot be opened for reading: {err!s}'.format(
            item=item, err=e
        ))
    return f

def get_user_choice(prompt, choices, case_lower=True):
    """ prompt user to make a choice. converts choices to lowercase unicode strings
        prompt: str:
        choices: list of str or int:
        case_lower: defaults to True for lower case enforcement
        return: unicode choice
    """
    if case_lower:
        _vals = [unicode(x).lower() for x in choices]
    else:
        _vals = [unicode(x) for x in choices]
    while True:
        if case_lower:
            _input = unicode(raw_input(prompt)).lower()
        else:
            _input = unicode(raw_input(prompt))
        if _input in _vals:
            return _input


#!/usr/bin/env python
# Copyright 2013-2017 Netflix, Inc.

from setuptools import setup, find_packages
from pip.req import parse_requirements
from pip.download import PipSession

setup(
    name = 'cligraphy',
    version = '0.0.1',
    description = 'Cligraphy Command line tools',
    long_description = 'Cligraphy Command line tools',
    author = 'Netflix, Inc.',
    author_email = '',  # OPEN SOURCE TODO
    packages = find_packages(),
    include_package_data=True,
    license='Apache 2.0',
    zip_safe=False,
    install_requires = [str(ir.req) for ir in parse_requirements('requirements.txt', session=PipSession())],
)



#!/usr/bin/env python
# Copyright 2015 Netflix, Inc.

"""
Wrappers and helpers around paramiko, with two general goals:
- concentrate dependency on paramiko in this module
- support our sometimes intricate uses cases (eg. efficient bastion hopping)
"""

from cligraphy.core import ctx

import paramiko
import OpenSSL.crypto as openssl

import os
import logging
import time

DEFAULT_TIMEOUT = 5
DEFAULT_COMMAND_TIMEOUT = 60

def _setup():
    # basically silence paramiko transport
    logging.getLogger('paramiko.transport').setLevel(logging.CRITICAL)


_setup()


class OpenSSLRSAKey(paramiko.rsakey.RSAKey):
    """For speed, use openssl instead of paramiko's slow pycrypto to sign our auth message"""

    def _from_private_key_file(self, filename, password):
        # paramiko key loading
        data = self._read_private_key_file('RSA', filename, password)
        self._decode_key(data)
        # openssl key loading
        with open(filename, 'rt') as fp:
            self._ssl_key = openssl.load_privatekey(openssl.FILETYPE_PEM, fp.read(), password)

    def sign_ssh_data(self, data):
        sig = openssl.sign(self._ssl_key, data, 'sha1')
        m = paramiko.message.Message()
        m.add_string('ssh-rsa')
        m.add_string(sig)
        return m

def get_proxy_client(password=None):
    """Creates a new client object connected to ${ssh.proxy.host}"""
    return ssh_client_builder().hostname(ctx.conf.ssh.proxy.host).password(password).allow_agent().ssh_config().connect()


def get_awsprod_client(password=None):
    """Creates a new client object connected to an awsprod bastion"""
    return ssh_client_builder().hostname('awsprod').password(password).allow_agent().ssh_config().connect()


def get_client(proxy_client, dest, username='root', port=22, agent=True, keys=None, password=None):
    """Creates a new client object connected to a remove server, by hopping through a bastion host using the given proxy_client"""

    if proxy_client:
        proxy_transport = proxy_client.get_transport()
        fwd = proxy_transport.open_channel("direct-tcpip", (dest, port), ('127.0.0.1', 0))
    else:
        fwd = None

    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    connected = False

    if not connected and agent:
        logging.debug('Trying agent auth')
        try:
            client.connect(dest, username=username, sock=fwd, timeout=DEFAULT_TIMEOUT, allow_agent=True)
            connected = True
        except paramiko.ssh_exception.SSHException:
            logging.debug('Could not auth using agent')


    if not connected and keys:
        logging.debug('Trying key auth')
        # FIXME: here we reconnect for each key attempt; don't do that and use the same connection instead.
        for key in keys:
            try:
                client.connect(dest, username=username, sock=fwd, timeout=DEFAULT_TIMEOUT, allow_agent=False, pkey=key)
                connected = True
                break
            except paramiko.ssh_exception.SSHException:
                continue

    if not connected and password:
        logging.debug('Trying password auth')
        try:
            client.connect(dest, username=username, sock=fwd, timeout=DEFAULT_TIMEOUT, allow_agent=False, password=password)
            connected = True
        except paramiko.ssh_exception.SSHException:
            logging.debug('Could not auth using password')

    if not connected:
        raise Exception('Could not authenticate')

    return client


class _SSHClientBuilder(object):

    def __init__(self):
        super(_SSHClientBuilder, self).__init__()
        self._hostname = None
        self._port = 22
        self._username = None
        self._password = None
        self._config = paramiko.SSHConfig()
        self._client_class = paramiko.SSHClient
        self._missing_host_key_policy = paramiko.AutoAddPolicy
        self._timeout = DEFAULT_TIMEOUT
        self._banner_timeout = DEFAULT_TIMEOUT
        self._allow_agent = None
        self._proxy_command = None
        self._sock = None

    def ssh_config(self, filename='~/.ssh/config'):
        """Parse the given ssh config file. Optional.
        No config file is parsed by default, but if this is called without a filename, ~/.ssh/config will be parsed."""
        absolute_filename = os.path.expanduser(filename)
        if not os.path.exists(absolute_filename):
            logging.warning('ssh config file %s does not exist, skipping it', absolute_filename)
            return self
        logging.debug('Reading ssh configuration from %s', absolute_filename)
        with open(absolute_filename, 'r') as fpin:
            self._config.parse(fpin)
        return self

    def client(self, client_class):
        """Sets the client class to build with. Optional, defaults to paramiko.SSHClient"""
        logging.debug('Setting client class to %s', client_class)
        self._client_class = client_class
        return self

    def missing_host_key_policy(self, policy_class):
        """Set the client class to build with. Optional, defaults to paramiko.AutoAddPolicy"""
        logging.debug('Setting missing host key policy to %s', policy_class)
        self._missing_host_key_policy = policy
        return self

    def hostname(self, hostname):
        """Set the destination hostname. Mandatory"""
        logging.debug('Setting hostname to %s', hostname)
        self._hostname = hostname
        return self

    def username(self, username):
        """Set the username. Optional, defaults to $USER"""
        logging.debug('Setting username to %s', username)
        self._username = username
        return self

    def password(self, password):
        """Set a password. Optional, no default"""
        logging.debug('Setting password (password not shown here for security reasons)')
        self._password = password
        return self

    def port(self, port):
        """Set the destination port number. Optional, defaults to 22"""
        logging.debug('Setting port to %d', port)
        self._port = port
        return self

    def timeout(self, timeout=5, banner_timeout=5):
        logging.debug('Setting timeouts: timeout=%d, banner_timeout=%d', timeout, banner_timeout)
        self._timeout = timeout
        self._banner_timeout = banner_timeout
        return self

    def allow_agent(self, allow_agent=True):
        """Allow or disallow ssh agent. Optional. Defaults to true"""
        self._allow_agent = allow_agent
        return self

    def proxy(self, proxy_command):
        self._proxy_command = proxy_command
        return self

    def sock(self, sock):
        self._sock = sock
        return self

    def connect(self):
        """Finish building the client and connects to the target server, returning a paramiko client object"""
        assert self._client_class
        assert self._hostname is not None, 'destination hostname was not specified'
        client = self._client_class()
        if self._missing_host_key_policy:
            client.set_missing_host_key_policy(self._missing_host_key_policy())

        config_data = self._config.lookup(self._hostname)
        ssh_kwargs = {
            'timeout': self._timeout,
            'banner_timeout': self._banner_timeout,
            'port': self._port
        }

        # unless one is explicitely specified with .user(), get our username from configuration, defaulting to $USER
        if self._username is None:
            ssh_kwargs['username'] = config_data.get('user', os.getenv('USER'))
        else:
            ssh_kwargs['username'] = self._username

        if self._password is not None:
            ssh_kwargs['password'] = self._password

        if 'proxycommand' in config_data:
            ssh_kwargs['sock'] = paramiko.ProxyCommand(config_data['proxycommand'])
        elif self._proxy_command is not None:
            ssh_kwargs['sock'] = paramiko.ProxyCommand(self._proxy_command)

        if config_data.get('identity_file') is not None:
            ssh_kwargs['key_filename'] = config_data.get('identity_file')

        # unless explicitely specified with .allow_agent, allow agent by default unless identitiesonly is yes in our config
        if self._allow_agent is None:
            ssh_kwargs['allow_agent'] = config_data.get('identitiesonly', 'no') != 'yes'
        else:
            ssh_kwargs['allow_agent'] = self._allow_agent

        if self._sock is not None:
            ssh_kwargs['sock'] = self._sock

        logging.debug('Connecting to %s with options %s', config_data['hostname'], ssh_kwargs)
        client.connect(config_data['hostname'], **ssh_kwargs)
        return client

def ssh_client_builder():
    return _SSHClientBuilder()


class TaskException(Exception):
    pass

class ConnectException(TaskException):
    pass

class CommandException(TaskException):
    pass


def exec_command(client, command, command_timeout=DEFAULT_COMMAND_TIMEOUT,
                 feed_data=None, get_files=[], put_files=[],
                 io_step_s=0.1, io_chunk_size=1024,
                 out=None, err=None):
    """Execute a command on the given client, grabbing stdout/stderr and the exit status code"""
    transport = client.get_transport()
    channel = transport.open_session()
    sftp = None
    if get_files or put_files:
        sftp = paramiko.SFTPClient.from_transport(transport)
        sftp.get_channel().settimeout(command_timeout)

    try:
        if put_files:
            for path, content in put_files.items():
                with sftp.open(path, 'wb') as remote_fp:
                    remote_fp.write(content)

        channel.exec_command(command)
        if feed_data is not None:
            channel.settimeout(command_timeout)
            channel.sendall(feed_data)

        channel.shutdown_write()
        channel.settimeout(io_step_s)
        out = out if out is not None else []
        err = err if err is not None else []
        deadline = time.time() + command_timeout
        while time.time() <= deadline and not channel.exit_status_ready():
            ready = channel.recv_ready(), channel.recv_stderr_ready()
            if not ready[0] and not ready[1]:
                time.sleep(io_step_s)
                continue
            if ready[0]:
                out.append(channel.recv(io_chunk_size))
            if ready[1]:
                err.append(channel.recv_stderr(io_chunk_size))

        if not channel.exit_status_ready():
            raise CommandException("Timeout during command execution")

        while channel.recv_ready():
            out.append(channel.recv(io_chunk_size))
        while channel.recv_stderr_ready():
            out.append(channel.recv_stderr(io_chunk_size))

        status = channel.recv_exit_status()

        get_files_result = {}
        if get_files:
            for path in get_files:
                with sftp.open(path, 'rb') as remote_fp:
                    get_files_result[path] = remote_fp.read()

        return out, err, status, get_files_result
    finally:
        if sftp:
            sftp.close()
        channel.close()
